import json
import os
import pathlib

from .nist_2011_atomic_weights import nist_2011_atomic_weights
from .nist_2014_codata import nist_2014_codata
from .alvarez_2008_covalent_radii import alvarez_2008_covalent_radii
